# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
This packages contains affiliated package tests.
"""
# basic unit tests of software functionality:
#from . import test_webbpsf

# more detailed validation against other people's sims:
from . import validate_webbpsf
from . import validate_vs_jwpsf
from . import test_webbpsf
from . import test_utils
